package ru.dbotthepony.kstarbound.defs.actor

import com.google.common.collect.ImmutableList
import com.google.gson.JsonObject
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.math.vector.Vector2d

@JsonFactory
data class HumanoidConfig(
	val globalOffset: Vector2d, // in pixels
	val headRunOffset: Vector2d, // in pixels
	val headSwimOffset: Vector2d, // in pixels
	val runFallOffset: Double, // in pixels
	val duckOffset: Double, // in pixels
	val headDuckOffset: Vector2d, // in pixels
	val sitOffset: Double, // in pixels
	val layOffset: Double, // in pixels
	val headSitOffset: Vector2d, // in pixels
	val headLayOffset: Vector2d, // in pixels
	val recoilOffset: Vector2d, // in pixels
	val mouthOffset: Vector2d, // in pixels
	val feetOffset: Vector2d, // in pixels

	val bodyFullbright: Boolean = false,

	val headArmorOffset: Vector2d, // in pixels
	val chestArmorOffset: Vector2d, // in pixels
	val legsArmorOffset: Vector2d, // in pixels
	val backArmorOffset: Vector2d, // in pixels

	val bodyHidden: Boolean = false,

	val armWalkSeq: ImmutableList<Int>,
	val armRunSeq: ImmutableList<Int>,

	val walkBob: ImmutableList<Double>, // in pixels
	val runBob: ImmutableList<Double>, // in pixels
	val swimBob: ImmutableList<Double>, // in pixels

	val jumpBob: Double,
	val frontArmRotationCenter: Vector2d, // in pixels
	val backArmRotationCenter: Vector2d, // in pixels
	val frontHandPosition: Vector2d, // in pixels
	val backArmOffset: Vector2d, // in pixels
	val vaporTrailFrames: Int,
	val vaporTrailCycle: Double,

	val deathParticles: String,
	val particleEmitters: JsonObject,

	val movementParameters: ActorMovementParameters,

	val personalities: ImmutableList<Personality>,
) {
	data class Timing(
		val stateCycle: ImmutableList<Double>? = null,
		val emoteCycle: ImmutableList<Double>? = null,
		val stateFrames: ImmutableList<Int>? = null,
		val emoteFrames: ImmutableList<Int>? = null,
	)
}