package ru.dbotthepony.kstarbound.lua

import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_NUMTYPES
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TBOOLEAN
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TFUNCTION
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TLIGHTUSERDATA
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TNIL
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TNONE
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TNUMBER
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TSTRING
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TTABLE
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TTHREAD
import ru.dbotthepony.kstarbound.lua.LuaThread.Companion.LUA_TUSERDATA

enum class LuaType {
	NONE {
		override val isNothing: Boolean
			get() = true
	},
	NIL {
		override val isNothing: Boolean
			get() = true
	},
	BOOLEAN {
		override val isBoolean: Boolean
			get() = true
	},
	LIGHTUSERDATA ,
	NUMBER {
		override val isNumber: Boolean
			get() = true
	},
	STRING {
		override val isString: Boolean
			get() = true
	},
	TABLE {
		override val isTable: Boolean
			get() = true
	},
	FUNCTION,
	USERDATA,
	THREAD {
		override val isThread: Boolean
			get() = true
	},
	UMTYPES;

	open val isNothing: Boolean
		get() = false

	open val isBoolean: Boolean
		get() = false

	open val isTable: Boolean
		get() = false

	open val isNumber: Boolean
		get() = false

	open val isString: Boolean
		get() = false

	open val isThread: Boolean
		get() = false

	val isBooleanOrNothing: Boolean get() = isBoolean || isNothing
	val isTableOrNothing: Boolean get() = isTable || isNothing
	val isNumberOrNothing: Boolean get() = isNumber || isNothing
	val isStringOrNothing: Boolean get() = isString || isNothing
	val isThreadOrNothing: Boolean get() = isThread || isNothing

	inline fun isBooleanOrNothing(formatter: LuaType.() -> String) {
		if (!isBooleanOrNothing) {
			throw IllegalArgumentException(formatter.invoke(this))
		}
	}

	inline fun isTableOrNothing(formatter: LuaType.() -> String) {
		if (!isTableOrNothing) {
			throw IllegalArgumentException(formatter.invoke(this))
		}
	}

	inline fun isNumberOrNothing(formatter: LuaType.() -> String) {
		if (!isNumberOrNothing) {
			throw IllegalArgumentException(formatter.invoke(this))
		}
	}

	inline fun isStringOrNothing(formatter: LuaType.() -> String) {
		if (!isStringOrNothing) {
			throw IllegalArgumentException(formatter.invoke(this))
		}
	}

	inline fun isThreadOrNothing(formatter: LuaType.() -> String) {
		if (!isThreadOrNothing) {
			throw IllegalArgumentException(formatter.invoke(this))
		}
	}

	fun isBooleanOrNothing(expected: String, argumentNumber: Int): Boolean {
		if (argumentNumber == -1) {
			return isBooleanOrNothing
		} else if (!isBooleanOrNothing) {
			throw IllegalArgumentException("bad argument $argumentNumber: $expected expected, got $this")
		}

		return isBooleanOrNothing
	}

	fun isTableOrNothing(expected: String, argumentNumber: Int): Boolean {
		if (argumentNumber == -1) {
			return isTableOrNothing
		} else if (!isTableOrNothing) {
			throw IllegalArgumentException("bad argument $argumentNumber: $expected expected, got $this")
		}

		return isTableOrNothing
	}

	fun isNumberOrNothing(expected: String, argumentNumber: Int): Boolean {
		if (argumentNumber == -1) {
			return isNumberOrNothing
		} else if (!isNumberOrNothing) {
			throw IllegalArgumentException("bad argument $argumentNumber: $expected expected, got $this")
		}

		return isNumberOrNothing
	}

	fun isStringOrNothing(expected: String, argumentNumber: Int): Boolean {
		if (argumentNumber == -1) {
			return isStringOrNothing
		} else if (!isStringOrNothing) {
			throw IllegalArgumentException("bad argument $argumentNumber: $expected expected, got $this")
		}

		return isStringOrNothing
	}

	fun isThreadOrNothing(expected: String, argumentNumber: Int): Boolean {
		if (argumentNumber == -1) {
			return isThreadOrNothing
		} else if (!isThreadOrNothing) {
			throw IllegalArgumentException("bad argument $argumentNumber: $expected expected, got $this")
		}

		return isThreadOrNothing
	}

	companion object {
		fun valueOf(value: Int): LuaType {
			return when (value) {
				LUA_TNONE -> NONE
				LUA_TNIL -> NIL
				LUA_TBOOLEAN -> BOOLEAN
				LUA_TLIGHTUSERDATA -> LIGHTUSERDATA
				LUA_TNUMBER -> NUMBER
				LUA_TSTRING -> STRING
				LUA_TTABLE -> TABLE
				LUA_TFUNCTION -> FUNCTION
				LUA_TUSERDATA -> USERDATA
				LUA_TTHREAD -> THREAD
				LUA_NUMTYPES -> UMTYPES
				else -> throw RuntimeException("Invalid Lua type: $value")
			}
		}
	}
}
