package ru.dbotthepony.kstarbound.lua.bindings

import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextAABB
import ru.dbotthepony.kstarbound.lua.nextColor
import ru.dbotthepony.kstarbound.lua.nextOptionalVector2f
import ru.dbotthepony.kstarbound.lua.nextPoly
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.nextVector2f
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.math.vector.Vector2f
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.entities.Animator

private fun setAnimationState(self: Animator, args: LuaThread.ArgStack): Int {
	val type = args.nextString()
	val state = args.nextString()
	val alwaysStart = args.nextOptionalBoolean() == true
	self.setActiveState(type, state, alwaysStart)
	return 0
}

private fun animationState(self: Animator, args: LuaThread.ArgStack): Int {
	val type = args.nextString()
	args.lua.push(self.animationState(type))
	return 1
}

private fun animationStateProperty(self: Animator, args: LuaThread.ArgStack): Int {
	val key = args.nextString()
	val value = args.nextString()
	args.lua.push(self.stateProperty(key, value))
	return 1
}

private fun setGlobalTag(self: Animator, args: LuaThread.ArgStack): Int {
	val key = args.nextString().sbIntern()
	val value = args.nextString().sbIntern()
	self.setGlobalTag(key, value)
	return 0
}

private fun setPartTag(self: Animator, args: LuaThread.ArgStack): Int {
	val part = args.nextString().sbIntern()
	val key = args.nextString().sbIntern()
	val value = args.nextString().sbIntern()
	self.setPartTag(part, key, value)
	return 0
}

private fun setFlipped(self: Animator, args: LuaThread.ArgStack): Int {
	self.isFlipped = args.nextBoolean()
	self.flippedRelativeCenterLine = args.nextOptionalDouble() ?: 0.0
	return 0
}

private fun setAnimationRate(self: Animator, args: LuaThread.ArgStack): Int {
	self.animationRate = args.nextDouble()
	return 0
}

private fun rotateGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	val rotation = args.nextDouble()
	val immediate = args.nextOptionalBoolean() == true
	self.rotateGroup(group, rotation, immediate)
	return 0
}

private fun currentRotationAngle(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	args.lua.push(self.currentRotationAngle(group))
	return 1
}

private fun targetRotationAngle(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(self.targetRotationAngle(args.nextString()))
	return 1
}

private fun has(self: Animator, getter: (Animator, String) -> Boolean, args: LuaThread.ArgStack): Int {
	args.lua.push(getter(self, args.nextString()))
	return 1
}

private fun pushNames(self: Animator, getter: (Animator) -> Collection<String>, args: LuaThread.ArgStack): Int {
	val values = getter(self)
	args.lua.pushTable(values.size)

	for ((i, v) in values.withIndex())
		args.lua.setTableValue(i + 1L, v)

	return 1
}

private fun translateTransformGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	val translation = args.nextVector2f()
	self.translateTransformGroup(group, translation)
	return 0
}

private fun rotateTransformGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	val rotation = args.nextFloat()
	val center = args.nextOptionalVector2f() ?: Vector2f.ZERO
	self.rotateTransformGroup(group, rotation, center)
	return 0
}

private fun scaleTransformationGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()

	if (args.peek() == LuaType.NUMBER) {
		val value = args.nextFloat()
		self.scaleTransformationGroup(group, Vector2f(value, value), args.nextOptionalVector2f() ?: Vector2f.ZERO)
	} else {
		self.scaleTransformationGroup(group, args.nextVector2f(), args.nextOptionalVector2f() ?: Vector2f.ZERO)
	}

	return 0
}

private fun transformTransformationGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	val r00 = args.nextFloat()
	val r01 = args.nextFloat()
	val r02 = args.nextFloat()
	val r10 = args.nextFloat()
	val r11 = args.nextFloat()
	val r12 = args.nextFloat()

	self.transformTransformationGroup(
		group,
		r00,
	    r01,
	    r02,
	    r10,
	    r11,
	    r12)

	return 0
}

private fun resetTransformationGroup(self: Animator, args: LuaThread.ArgStack): Int {
	val group = args.nextString()
	self.resetTransformationGroup(group)
	return 0
}

private fun setParticleEmitterActive(self: Animator, args: LuaThread.ArgStack): Int {
	self.setParticleEmitterActive(args.nextString(), args.nextBoolean())
	return 0
}

private fun setParticleEmitterEmissionRate(self: Animator, args: LuaThread.ArgStack): Int {
	self.setParticleEmitterEmissionRate(args.nextString(), args.nextDouble())
	return 0
}

private fun setParticleEmitterBurstCount(self: Animator, args: LuaThread.ArgStack): Int {
	self.setParticleEmitterBurstCount(args.nextString(), args.nextInt())
	return 0
}

private fun setParticleEmitterOffsetRegion(self: Animator, args: LuaThread.ArgStack): Int {
	self.setParticleEmitterOffsetRegion(args.nextString(), args.nextAABB())
	return 0
}

private fun burstParticleEmitter(self: Animator, args: LuaThread.ArgStack): Int {
	self.burstParticleEmitter(args.nextString())
	return 0
}

private fun setLightActive(self: Animator, args: LuaThread.ArgStack): Int {
	self.setLightActive(args.nextString(), args.nextBoolean())
	return 0
}

private fun setLightPosition(self: Animator, args: LuaThread.ArgStack): Int {
	self.setLightPosition(args.nextString(), args.nextVector2d())
	return 0
}

private fun setLightColor(self: Animator, args: LuaThread.ArgStack): Int {
	self.setLightColor(args.nextString(), args.nextColor())
	return 0
}

private fun setLightPointAngle(self: Animator, args: LuaThread.ArgStack): Int {
	self.setLightPointAngle(args.nextString(), args.nextDouble())
	return 0
}

private fun setSoundPool(self: Animator, args: LuaThread.ArgStack): Int {
	val pool = args.nextString()
	val sounds = args.readTableValues { getString(it)?.sbIntern() ?: throw IllegalArgumentException("invalid values in sound pool table") }
	self.setSoundPool(pool, sounds)
	return 0
}

private fun setSoundPosition(self: Animator, args: LuaThread.ArgStack): Int {
	self.setSoundPosition(args.nextString(), args.nextVector2d())
	return 0
}

private fun playSound(self: Animator, args: LuaThread.ArgStack): Int {
	self.playSound(args.nextString(), args.nextOptionalInt() ?: 0)
	return 0
}

private fun setSoundVolume(self: Animator, args: LuaThread.ArgStack): Int {
	self.setSoundVolume(args.nextString(), args.nextDouble(), args.nextOptionalDouble() ?: 0.0)
	return 0
}

private fun setSoundPitch(self: Animator, args: LuaThread.ArgStack): Int {
	self.setSoundPitch(args.nextString(), args.nextDouble(), args.nextOptionalDouble() ?: 0.0)
	return 0
}

private fun stopAllSounds(self: Animator, args: LuaThread.ArgStack): Int {
	self.stopAllSounds(args.nextString(), args.nextOptionalDouble() ?: 0.0)
	return 0
}

private fun setEffectActive(self: Animator, args: LuaThread.ArgStack): Int {
	self.setEffectActive(args.nextString(), args.nextBoolean())
	return 0
}

private fun partPoint(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(self.partPoint(args.nextString(), args.nextString()))
	return 1
}

private fun partPoly(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(self.partPoly(args.nextString(), args.nextString()))
	return 1
}

private fun partProperty(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(self.partProperty(args.nextString(), args.nextString()))
	return 1
}

private fun transformPoint(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(args.nextVector2d() * self.partTransformation(args.nextString()))
	return 1
}

private fun transformPoly(self: Animator, args: LuaThread.ArgStack): Int {
	args.lua.push(args.nextPoly().transform(self.partTransformation(args.nextString())))
	return 1
}

fun provideAnimatorBindings(self: Animator, lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("animator")

	lua.pushBinding(self, "setAnimationState", ::setAnimationState)
	lua.pushBinding(self, "animationState", ::animationState)
	lua.pushBinding(self, "animationStateProperty", ::animationStateProperty)
	lua.pushBinding(self, "setGlobalTag", ::setGlobalTag)
	lua.pushBinding(self, "setPartTag", ::setPartTag)
	lua.pushBinding(self, "setFlipped", ::setFlipped)
	lua.pushBinding(self, "setAnimationRate", ::setAnimationRate)
	lua.pushBinding(self, "rotateGroup", ::rotateGroup)
	lua.pushBinding(self, "currentRotationAngle", ::currentRotationAngle)
	lua.pushBinding(self, "targetRotationAngle", ::targetRotationAngle)
	lua.pushBinding(self, "translateTransformGroup", ::translateTransformGroup)
	lua.pushBinding(self, "rotateTransformGroup", ::rotateTransformGroup)
	lua.pushBinding(self, "scaleTransformationGroup", ::scaleTransformationGroup)
	lua.pushBinding(self, "transformTransformationGroup", ::transformTransformationGroup)
	lua.pushBinding(self, "resetTransformationGroup", ::resetTransformationGroup)
	lua.pushBinding(self, "setParticleEmitterActive", ::setParticleEmitterActive)
	lua.pushBinding(self, "setParticleEmitterEmissionRate", ::setParticleEmitterEmissionRate)
	lua.pushBinding(self, "setParticleEmitterBurstCount", ::setParticleEmitterBurstCount)
	lua.pushBinding(self, "setParticleEmitterOffsetRegion", ::setParticleEmitterOffsetRegion)
	lua.pushBinding(self, "burstParticleEmitter", ::burstParticleEmitter)
	lua.pushBinding(self, "setLightActive", ::setLightActive)
	lua.pushBinding(self, "setLightPosition", ::setLightPosition)
	lua.pushBinding(self, "setLightColor", ::setLightColor)
	lua.pushBinding(self, "setLightPointAngle", ::setLightPointAngle)
	lua.pushBinding(self, "setSoundPool", ::setSoundPool)
	lua.pushBinding(self, "setSoundPosition", ::setSoundPosition)
	lua.pushBinding(self, "playSound", ::playSound)
	lua.pushBinding(self, "setSoundVolume", ::setSoundVolume)
	lua.pushBinding(self, "setSoundPitch", ::setSoundPitch)
	lua.pushBinding(self, "stopAllSounds", ::stopAllSounds)
	lua.pushBinding(self, "setEffectActive", ::setEffectActive)
	lua.pushBinding(self, "partPoint", ::partPoint)
	lua.pushBinding(self, "partPoly", ::partPoly)
	lua.pushBinding(self, "partProperty", ::partProperty)
	lua.pushBinding(self, "transformPoint", ::transformPoint)
	lua.pushBinding(self, "transformPoly", ::transformPoly)

	lua.pushBinding(self, Animator::hasRotationGroup, "hasRotationGroup", ::has)
	lua.pushBinding(self, Animator::hasTransformationGroup, "hasTransformationGroup", ::has)
	lua.pushBinding(self, Animator::hasParticleEmitter, "hasParticleEmitter", ::has)
	lua.pushBinding(self, Animator::hasLight, "hasLight", ::has)
	lua.pushBinding(self, Animator::hasSound, "hasSound", ::has)
	lua.pushBinding(self, Animator::hasEffect, "hasEffect", ::has)
	lua.pushBinding(self, Animator::hasPart, "hasPart", ::has)

	lua.pushBinding(self, Animator::rotationGroups, "rotationGroups", ::pushNames)
	lua.pushBinding(self, Animator::transformationGroups, "transformationGroups", ::pushNames)
	lua.pushBinding(self, Animator::particleEmitters, "particleEmitters", ::pushNames)
	lua.pushBinding(self, Animator::lights, "lights", ::pushNames)
	lua.pushBinding(self, Animator::sounds, "sounds", ::pushNames)
	lua.pushBinding(self, Animator::effects, "effects", ::pushNames)
	lua.pushBinding(self, Animator::parts, "parts", ::pushNames)

	lua.pop()
}
