package ru.dbotthepony.kstarbound.lua.bindings

import com.google.common.collect.ImmutableList
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.quest.QuestArcDescriptor
import ru.dbotthepony.kstarbound.defs.quest.QuestDescriptor
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.util.sbIntern

fun <T : Any> setOfferedQuests(self: T, prop: (T) -> MutableList<QuestArcDescriptor>, args: LuaThread.ArgStack): Int {
	val offeredQuests = prop(self)
	offeredQuests.clear()

	if (args.hasNext) {
		offeredQuests.addAll(args.readTableValues {
			val peek = typeAt(it)

			if (peek == LuaType.TABLE) {
				Starbound.gson.fromJson(getJson(it)!!, QuestArcDescriptor::class.java)
			} else if (peek == LuaType.STRING) {
				QuestArcDescriptor(ImmutableList.of(QuestDescriptor(getString(it)!!.sbIntern())))
			} else {
				throw IllegalArgumentException("Invalid value in provided quests table: $peek")
			}
		})
	}

	return 0
}

fun <T : Any> setTurnInQuests(self: T, prop: (T) -> MutableList<String>, args: LuaThread.ArgStack): Int {
	val turnInQuests = prop(self)
	turnInQuests.clear()

	if (args.hasNext) {
		turnInQuests.addAll(args.readTableValues { getString(it)?.sbIntern() ?: throw IllegalArgumentException("Turn-in quests table contains non-string values") })
	}

	return 0
}
