package ru.dbotthepony.kstarbound.lua.bindings

import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.math.Line2d
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.ActorEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity
import ru.dbotthepony.kstarbound.world.entities.NPCEntity
import ru.dbotthepony.kstarbound.world.entities.player.PlayerEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject

private fun id(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.entityID.toLong())
	return 1
}

private fun position(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.position)
	return 1
}

private fun entityType(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.type.jsonName)
	return 1
}

private fun uniqueId(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.uniqueID.get())
	return 1
}

private fun persistent(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.isPersistent)
	return 1
}

private fun entityInSight(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	val target = self.world.entities[args.nextInt()]

	if (target == null) {
		args.lua.push(false)
	} else {
		args.lua.push(self.world.chunkMap.collide(Line2d(self.position, target.position)) { it.type.isSolidCollision } == null)
	}

	return 1
}

private fun isValidTarget(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	val target = self.world.entities[args.nextInt()]

	if (target == null) {
		args.lua.push(false)
	} else {
		if (!self.team.get().canDamage(target.team.get(), self == target)) // original engine always passes `false` to `isSelfDamage`
			args.lua.push(false)
		else if (target is MonsterEntity)
			args.lua.push(target.isAggressive)
		else
			args.lua.push(target is PlayerEntity)

		// TODO: NPC handling here
	}

	return 1
}

private fun damageTeam(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	args.lua.pushTable(hashSize = 2)
	args.lua.setTableValue("team", self.team.get().team)
	args.lua.setTableValue("type", self.team.get().type.jsonName)
	return 1
}

private fun distanceToEntity(self: AbstractEntity, args: LuaThread.ArgStack): Int {
	val target = self.world.entities[args.nextInt()]

	if (target != null) {
		args.lua.push(self.world.geometry.diff(target.position, self.position))
	} else {
		args.lua.push(Vector2d.ZERO)
	}

	return 1
}

fun provideEntityBindings(self: AbstractEntity, lua: LuaThread) {
	provideWorldBindings(self.world, lua)

	if (self is WorldObject)
		provideWorldObjectBindings(self, lua)

	if (self is MonsterEntity)
		provideMonsterBindings(self, lua)

	if (self is ActorEntity)
		provideStatusControllerBindings(self.statusController, lua)

	if (self is NPCEntity)
		provideNPCBindings(self, lua)

	lua.pushTable()
	lua.dup()
	lua.storeGlobal("entity")

	lua.pushBinding(self, "id", ::id)
	lua.pushBinding(self, "position", ::position)
	lua.pushBinding(self, "entityType", ::entityType)
	lua.pushBinding(self, "uniqueId", ::uniqueId)
	lua.pushBinding(self, "persistent", ::persistent)
	lua.pushBinding(self, "entityInSight", ::entityInSight)
	lua.pushBinding(self, "isValidTarget", ::isValidTarget)
	lua.pushBinding(self, "damageTeam", ::damageTeam)
	lua.pushBinding(self, "distanceToEntity", ::distanceToEntity)

	lua.pop()
}
