package ru.dbotthepony.kstarbound.lua.userdata

import ru.dbotthepony.kstarbound.lua.LuaHandle
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.setTableValue
import ru.dbotthepony.kstarbound.util.CarriedExecutor
import ru.dbotthepony.kstarbound.util.supplyAsync
import ru.dbotthepony.kstarbound.world.entities.PathFinder

class LuaPathFinder(private val pacer: CarriedExecutor, private val self: PathFinder) {
	private var isRunning = false

	fun result(args: LuaThread.ArgStack): Int {
		if (self.result.isEmpty)
			return 0

		val list = self.result.value ?: return 0
		return convertPath(args.lua, list)
	}

	fun explore(args: LuaThread.ArgStack): Int {
		val maxExplore = args.nextOptionalLong()?.toInt()
		val useOffThread = args.nextOptionalBoolean() ?: true

		if (isRunning) {
			if (self.result.isEmpty)
				return 0
			else {
				args.lua.push(self.result.value != null)
				return 1
			}
		} else if (self.result.isPresent) {
			args.lua.push(self.result.value != null)
			return 1
		} else {
			val immediateMaxExplore = maxExplore?.times(4)?.coerceAtMost(800) ?: 800
			val result = self.run(immediateMaxExplore)

			if (result != null) {
				args.lua.push(result)
				return 1
			} else if (self.result.isEmpty && useOffThread) {
				// didn't explore enough, run in separate thread to not block main thread
				pacer.supplyAsync(self)
				isRunning = true
				return 0
			}
		}

		return 0
	}

	fun isExploringOffThread(args: LuaThread.ArgStack): Int {
		args.lua.push(isRunning && self.result.isEmpty)
		return 1
	}

	fun usedOffThread(args: LuaThread.ArgStack): Int {
		args.lua.push(isRunning)
		return 1
	}

	fun runAsync(args: LuaThread.ArgStack): Int {
		if (!isRunning) {
			pacer.supplyAsync(self)
			isRunning = true
		}

		return 0
	}

	companion object {
		fun convertPath(lua: LuaThread, list: List<PathFinder.Edge>?): Int {
			list ?: return 0
			lua.pushTable(list.size)
			var i = 1L

			for (edge in list) {
				lua.push(i++)

				lua.pushTable(hashSize = 5)

				lua.setTableValue("cost", edge.cost)
				lua.setTableValue("action", edge.action.jsonName)
				lua.setTableValue("jumpVelocity", edge.velocity)

				lua.push("source")
				edge.source.store(lua)
				lua.setTableValue()

				lua.push("target")
				edge.target.store(lua)
				lua.setTableValue()

				lua.setTableValue()
			}

			return 1
		}

		fun initializeHandle(lua: LuaThread): LuaHandle {
			lua.pushTable()
			val handle = lua.createHandle("PathFinder")

			lua.pushBinding("explore", LuaPathFinder::explore)
			lua.pushBinding("result", LuaPathFinder::result)
			lua.pushBinding("runAsync", LuaPathFinder::runAsync)
			lua.pushBinding("isExploringOffThread", LuaPathFinder::isExploringOffThread)
			lua.pushBinding("usedOffThread", LuaPathFinder::usedOffThread)

			lua.pop()

			return handle
		}
	}
}

fun LuaThread.push(value: LuaPathFinder) {
	pushTable()

	push("__index")
	push(commonHandles.pathFinder)
	setTableValue()

	pushObject(value)
}
