package ru.dbotthepony.kstarbound.lua.userdata

import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.util.random.nextNormalDouble
import ru.dbotthepony.kstarbound.util.random.random
import java.util.random.RandomGenerator

class LuaRandom(var random: RandomGenerator) {
	fun init(args: LuaThread.ArgStack): Int {
		random = random(args.nextLong())
		return 0
	}

	fun addEntropy(args: LuaThread.ArgStack): Int {
		throw UnsupportedOperationException("Adding entropy is not supported on new engine. If you have legitimate usecase for this, please let us know on issue tracker or discord")
	}

	fun randu32(args: LuaThread.ArgStack): Int {
		args.lua.push(random.nextLong(0, Int.MAX_VALUE.toLong()))
		return 1
	}

	fun randi32(args: LuaThread.ArgStack): Int {
		args.lua.push(random.nextLong(Int.MIN_VALUE.toLong(), Int.MAX_VALUE.toLong()))
		return 1
	}

	fun randu64(args: LuaThread.ArgStack): Int {
		args.lua.push(random.nextLong(0, Long.MAX_VALUE))
		return 1
	}

	fun randi64(args: LuaThread.ArgStack): Int {
		args.lua.push(random.nextLong(Long.MIN_VALUE, Long.MAX_VALUE))
		return 1
	}

	fun randf(args: LuaThread.ArgStack): Int {
		val origin = args.nextOptionalDouble()?.toFloat()
		val bound = args.nextOptionalDouble()?.toFloat()

		if (origin == null || bound == null) {
			args.lua.push(random.nextFloat())
		} else {
			require(origin <= bound) { "interval is empty: $origin <= $bound" }

			if (origin == bound) {
				// old behavior where it still updates internal random generator state
				random.nextFloat()
				args.lua.push(origin)
			} else
				args.lua.push(random.nextFloat(origin, bound))
		}

		return 1
	}

	fun randd(args: LuaThread.ArgStack): Int {
		val origin = args.nextOptionalDouble()
		val bound = args.nextOptionalDouble()

		if (origin == null || bound == null) {
			args.lua.push(random.nextDouble())
		} else {
			require(origin <= bound) { "interval is empty: $origin <= $bound" }

			if (origin == bound) {
				// old behavior where it still updates internal random generator state
				random.nextDouble()
				args.lua.push(origin)
			} else
				args.lua.push(random.nextDouble(origin, bound))
		}

		return 1
	}

	fun randLong(args: LuaThread.ArgStack): Int {
		val origin = args.nextLong()
		val bound = args.nextOptionalLong()

		if (bound == null) {
			if (origin == 0L) {
				random.nextLong() // to keep old behavior
				args.lua.push(0L)
			} else if (origin < 0L) {
				args.lua.push(random.nextLong(origin, 0L))
			} else {
				args.lua.push(random.nextLong(0L, origin))
			}
		} else {
			if (bound == origin) {
				// old behavior where it still updates internal random generator state
				random.nextLong()
				args.lua.push(origin)
			} else {
				args.lua.push(random.nextLong(origin, bound))
			}
		}

		return 1
	}

	fun randn(args: LuaThread.ArgStack): Int {
		args.lua.push(random.nextNormalDouble(args.nextDouble(), args.nextDouble()))
		return 1
	}
}
