package ru.dbotthepony.kstarbound.network.packets.serverbound

import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import ru.dbotthepony.kstarbound.world.entities.tile.WireConnection
import java.io.DataInputStream
import java.io.DataOutputStream

class ConnectWirePacket(val target: WireConnection, val source: WireConnection) : IServerPacket {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this(WireConnection(stream), WireConnection(stream))

	override fun write(stream: DataOutputStream, isLegacy: Boolean) {
		target.write(stream)
		source.write(stream)
	}

	override suspend fun play(connection: ServerConnection) {
		connection.enqueueAndSuspend {
			val target = entityIndex.tileEntityAt(target.entityLocation, WorldObject::class) ?: return@enqueueAndSuspend
			val source = entityIndex.tileEntityAt(source.entityLocation, WorldObject::class) ?: return@enqueueAndSuspend

			val targetNode = target.outputNodes.getOrNull(this@ConnectWirePacket.target.index) ?: return@enqueueAndSuspend
			val sourceNode = source.inputNodes.getOrNull(this@ConnectWirePacket.source.index) ?: return@enqueueAndSuspend

			if (this@ConnectWirePacket.source in targetNode.connections && this@ConnectWirePacket.target in sourceNode.connections) {
				// disconnect
				targetNode.removeConnection(this@ConnectWirePacket.source)
				sourceNode.removeConnection(this@ConnectWirePacket.target)
			} else {
				targetNode.addConnection(this@ConnectWirePacket.source)
				sourceNode.addConnection(this@ConnectWirePacket.target)
			}
		}
	}
}
